/***************************************************************************
 *   Copyright (C) 2006 - 2007 by Ian Ventura-Whiting (Fizz)               *
 *   Fizz@Titania.co.uk                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************

The following files should be provided with this source code:

* INSTALL   - The compilation and installation guide
* LICENCE   - The licence covering all the files
* Changelog - A listing of all source code changes
* Readme    - Information about Cisco Parse


Program Overview
~~~~~~~~~~~~~~~~

Nipper is a Network Infrastructure Configuration Parser. Nipper takes
a network infrastructure device configuration, processes the file and
details security-related issues with the configuration together with
detailed recommendations. Nipper was previous known as CiscoParse.

Nipper currently supports Cisco switches (IOS), Cisco Routers (IOS),
Cisco Firewalls (PIX/ASA/FWSM) and Juniper NetScreen (ScreenOS). Output
is in HTML, Latex, XML and Text. Encrypted passwords can be output to a
John-the-Ripper file for strength testing.


Program Usage
~~~~~~~~~~~~~

For parsing a Cisco router configuration file called router.txt and
outputting as a text file called report.txt:

$ nipper --input=router.txt --output=report.txt --output-format=text

For parsing a PIX firewall configuration file called firewall.txt and
outputing as a HTML file called report.html:

$ nipper --pix --input=firewall.txt --output=report.html

For more options, view the nipper help:

$ nipper --help


Program Code
~~~~~~~~~~~~

This program code is split into five distinct parts:

  1) Program parameters and initialisation.
  2) Process the configuration file.
  3) Report Pre-processing (defaults...).
  4) Reporting and output
  5) Cleanup

*/


// Standard Includes...
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

// Global Constants...
const char *program_banner = "                   _                           ____\n             _ __ (_)_ __  _ __   ___ _ __    / ->/|\n            | '_ \\| | '_ \\| '_ \\ / _ \\ '__|  /<-_/ |\n            | | | | | |_) | |_) |  __/ |     |   | /\n            |_| |_|_| .__/| .__/ \\___|_|     |___|/\n                    |_|   |_|\n\n                        Version 0.9.5\n                by Ian Ventura-Whiting (Fizz)\n";

// Common structure include...
#include "nipper-struct.c"

// Common functions include...
#include "nipper-common.c"

// Common report include...
#include "report-appendix.c"
#include "report-text.c"
#include "report-functions.c"

// Processing IOS includes...
#include "IOS/process-aaa.c"
#include "IOS/process-access-list.c"
#include "IOS/process-banner.c"
#include "IOS/process-enable.c"
#include "IOS/process-ftp.c"
#include "IOS/process-general.c"
#include "IOS/process-http.c"
#include "IOS/process-interface.c"
#include "IOS/process-keychain.c"
#include "IOS/process-line.c"
#include "IOS/process-logging.c"
#include "IOS/process-ntp.c"
#include "IOS/process-router.c"
#include "IOS/process-snmp.c"
#include "IOS/process-ssh.c"
#include "IOS/process-tacacs.c"
#include "IOS/process-username.c"

// Report IOS includes...
#include "IOS/report-passwords.c"
#include "IOS/report-timeout.c"
#include "IOS/report-debug.c"
#include "IOS/report.c"

// Processing NMP includes...
#include "NMP/process-general.c"
#include "NMP/process-permit.c"
#include "NMP/process-port.c"
#include "NMP/process-snmp.c"
#include "NMP/process-spantree.c"
#include "NMP/process-vlan.c"
#include "NMP/process-cdp.c"

// Report NMP includes...
#include "NMP/report-passwords.c"
#include "NMP/report-debug.c"
#include "NMP/report.c"

// Processing PIX includes...
#include "PIX/process-access-list.c"
#include "PIX/process-enable.c"
#include "PIX/process-general.c"
#include "PIX/process-icmp.c"
#include "PIX/process-interface.c"
#include "PIX/process-nat.c"
#include "PIX/process-objects.c"
#include "PIX/process-protocols.c"
#include "PIX/process-route.c"
#include "PIX/process-ssh.c"

// Report PIX includes...
#include "PIX/report-debug.c"
#include "PIX/report-access-list.c"
#include "PIX/report.c"

// Processing ScreenOS includes...
#include "ScreenOS/process-address.c"
#include "ScreenOS/process-admin.c"
#include "ScreenOS/process-auth-server.c"
#include "ScreenOS/process-interface.c"
#include "ScreenOS/process-policy.c"
#include "ScreenOS/process-ssh.c"

// Report SOS includes...
#include "ScreenOS/report-debug.c"
#include "ScreenOS/report.c"


// ----------------------------------------------------
// PART 1: Program parameters & initialization

int main(int argc, char *argv[])
{
	// Program variables...
	int showUsage = false;			// Should usage information be shown? (default to false)
	int inputFile = 0;				// Argument that contains input file, if specified.
	char outputFile[128];			// Argument that contains output file, if specified.
	int outputFormat = 1;			// Output file format output_text...
	int debugMode = false;			// Show debug output
	int argLoop;					// Arguement loop variable
	int passwords = true;			// Output passwords?
	int tempInt;
	int tempInt2;
	int linesnotprocessed;			// Print the lines that were not processed (true or false)
	char john[128];					// John the ripper file name
	char companyName[128];
	FILE *input;					// Input file
	char line[1024];				// Line from configuration file
	FILE *dictionaryFile;			// Dictionary file
	FILE *cssFile;					// HTML CSS file
	int abs = true;					// Abbreviation output
	int appendix = true;			// Appendix output
	int glossary = true;			// Glossary output
	int commonPorts = true;			// Common Ports output
	int fullReports = true;			// Output full reports
	int debugAbbrev = false;		// Output Abbreviations that were not found
	int expandACL = false;			// Expand an ACL's object definitions.
	int names = true;				// Show hostnames instead of IP addresses (where possible)
	int deviceType;					// Type of Device (switch, firewall...)
	char remoteIP[16];				// Retrival of Config remotely - IP
	char localIP[16];				// Retrival of Config remotely - IP (local)
	char remoteTftp[128];			// Retrival of Config remotely - TFTP root
	char remoteSnmp[128];			// Retrival of Config remotely - SNMP community
	char tempCommand[256];			// Retrival of Config remotely - Command to execute
	int localSave;					// Retrival of Config remotely - Config Save Name
	int oldCisco = false;			// Retrival of Config remotely - Old or new methods	
	struct stat *fileStats;
	struct ciscoCommand command;
	struct latexConfig latex;
	struct debugAbbrev *debugAbPointer;

	// IOS Configuration variables...
	struct ciscoIOSConfig configIOS;
	struct accessList *accessPointer;
	struct accessListPermission *permissionPointer;
	struct enablePassword *enablePointer;
	struct ntpKey *ntpKeyPointer;
	struct ntpHost *ntpHostPointer;
	struct ntpAccessGroup *ntpAccessPointer;
	struct snmpCommunity *communityPointer;
	struct snmpViewItem *snmpViewItemPointer;
	struct snmpView *snmpViewPointer;
	struct snmpGroup *snmpGroupPointer;
	struct snmpUser *snmpUserPointer;
	struct snmpTrap *snmpTrapPointer;
	struct snmpHost *snmpHostPointer;
	struct usernameList *usernamePointer;
	struct privilageLevels *privilagePointer;
	struct host *logServerPointer;
	struct host *tacacsHostPointer;
	struct aaaGroupServer *aaaGroupServerPointer;
	struct aaaServer *aaaServerPointer;
	struct netInterface *interfacePointer;
	struct interfaceConfig *interfaceConfigPointer;
	struct bgpNeighbor *neighborPointer;
	struct keyChain *keyChainPointer;
	struct keyConfig *keyPointer;
	struct lineConfig *linePointer;
	struct bannerConfig *bannerPointer;
	struct ospfNetworkConfig *ospfNetworkPointer;
	struct ospfAreaConfig *ospfAreaPointer;
	struct ospfNeighbor *ospfNeighborPointer;
	struct staticRoute *routePointer;

	// PIX Configuration variables...
	struct ciscoPIXConfig configPIX;
	struct interfacePIXConfig *interfacePIXPointer;
	struct accessListPIX *accessPIXPointer;
	struct accessListPIXPermission *permissionPIXPointer;
	struct objectGroupList *objectGroupPointer;
	struct objectList *objectPointer;
	struct namePIXList *namePIXPointer;
	struct enablePIXPassword *enablePIXPointer;
	struct usernamePIX *usernamePIXPointer;
	struct fixupList *fixupPIXPointer;
	struct icmpAccess *icmpPIXPointer;
	struct staticPIX *staticPIXPointer;
	struct routePIX *routePIXPointer;

	// NMP Configuration variables...
	struct ciscoNMPConfig configNMP;
	struct snmpNMPCommunity *communityNMPPointer;
	struct snmpNMPHost *snmpNMPHostPointer;
	struct hostNMP *hostNMPPointer;
	struct vlanConfig *vlanNMPPointer;
	struct moduleConfig *moduleNMPPointer;
	struct portConfig *portNMPPointer;

	// ScreenOS Configuration variables...
	struct screenOSConfig configSOS;
	struct adminUsersSOS *userSOSPointer;
	struct authServerSOS *authServerPointer;
	struct interfaceSOS *interfaceSOSPointer;
	struct policySOS *policySOSPointer;
	struct zonePolicyRuleSOS *ruleSOSPointer;
	struct listSOS *listSOSPointer;
	struct addressSOS *addressSOSPointer;
	struct mappingSOS *mappingSOSPointer;

	// Combined Configuration Variables...
	struct johnPassword *johnPointer;

	// Init program configuration...
	memset(john, 0, sizeof(john));
	memset(outputFile, 0, sizeof(outputFile));
	memset(remoteIP, 0, sizeof(remoteIP));
	memset(localIP, 0, sizeof(localIP));
	memset(remoteSnmp, 0, sizeof(remoteSnmp));
	strcpy(remoteSnmp, "private");
	memset(remoteTftp, 0, sizeof(remoteTftp));
	strcpy(remoteTftp, "/tftproot/");
	memset(tempCommand, 0, sizeof(tempCommand));
	memset(companyName, 0, sizeof(companyName));
	strcpy(companyName, "Nipper");
	memset(&latex, 0, sizeof(struct latexConfig));
	strcpy(latex.paperSize, "a4paper");
	strcpy(latex.documentClass, "article");
	localSave = 0;
	dictionaryFile = 0;
	cssFile = 0;
	fullReports = true;
	deviceType = type_router;
	linesnotprocessed = false;

	// Init IOS configuration...
	memset(&configIOS, 0, sizeof(struct ciscoIOSConfig));
	configIOS.sourceRouting = true;
	configIOS.tcpSmallServers = unconfigured;
	configIOS.udpSmallServers = unconfigured;
	configIOS.finger = unconfigured;
	configIOS.cdp = true;
	configIOS.tftp = unconfigured;
	configIOS.telnet = unconfigured;
	configIOS.servicePasswordEncryption = false;
	configIOS.bootp = true;
	configIOS.bootNetwork = false;
	configIOS.serviceConfig = false;
	configIOS.pad = true;
	configIOS.domainLookup = true;

	// Init NMP configuration
	memset(&configNMP, 0, sizeof(struct ciscoNMPConfig));
	configNMP.cdp = true;
	configNMP.cdpInterval = 60;
	configNMP.cdpHoldtime = 180;
	configNMP.cdpVersion = 2;
	configNMP.timeout = 20;
	configNMP.portSecurityAuto = false;

	// Init PIX configuration...
	memset(&configPIX, 0, sizeof(struct ciscoPIXConfig));
	configPIX.type = firewall_type_pix;
	configPIX.transparent = false;
	configPIX.accessListAlertInterval = 300;
	configPIX.accessListDenyFlowMax = 4096;

	// Init ScreenOS configuration
	memset(&configSOS, 0, sizeof(struct screenOSConfig));
	configSOS.sshEnabled = false;
	configSOS.sshKeyGenTime = 60;

	// Get program parameters
	for (argLoop = 1; argLoop < argc; argLoop++)
	{
		// Output format specified, defaults to text
		if ((strncmp("--output-format=", argv[argLoop], 16) == 0) || (strncmp("--report-format=", argv[argLoop], 16) == 0))
		{
			if (strcmp("text", argv[argLoop]+16) == 0)
				outputFormat = output_text;
			else if (strcmp("html", argv[argLoop]+16) == 0)
				outputFormat = output_html;
			else if (strcmp("latex", argv[argLoop]+16) == 0)
				outputFormat = output_latex;
			else if (strcmp("xml", argv[argLoop]+16) == 0)
				outputFormat = output_xml;
			else
				showUsage = true;
		}
		else if (strncmp("--text", argv[argLoop], 6) == 0)
			outputFormat = output_text;
		else if (strncmp("--html", argv[argLoop], 6) == 0)
			outputFormat = output_html;
		else if (strncmp("--latex", argv[argLoop], 7) == 0)
			outputFormat = output_latex;
		else if (strncmp("--xml", argv[argLoop], 5) == 0)
			outputFormat = output_xml;

		// Input file (defaults to stdin)
		else if ((strncmp("--input=", argv[argLoop], 8) == 0) && (strlen(argv[argLoop]) > 8))
			inputFile = argLoop;

		// Output file (defaults to stdout)
		else if (((strncmp("--output=", argv[argLoop], 9) == 0) || (strncmp("--report=", argv[argLoop], 9) == 0)) && (strlen(argv[argLoop]) > 9))
			strncpy(outputFile, argv[argLoop]+9, sizeof(outputFile) - 1);

		// Password output
		else if (strcmp("--no-passwords", argv[argLoop]) == 0)
			passwords = false;

		// Do not perform a security audit
		else if (strcmp("--no-audit", argv[argLoop]) == 0)
			fullReports = false;

		// Expand ACL object definitions
		else if (strcmp("--expand-acl", argv[argLoop]) == 0)
			expandACL = true;

		// Debug output
		else if (strcmp("--debug", argv[argLoop]) == 0)
			debugMode = true;
		else if (strcmp("--debug-lnp", argv[argLoop]) == 0)
			linesnotprocessed = true;

		// Router
		else if ((strcmp("--router", argv[argLoop]) == 0) || (strcmp("--ios", argv[argLoop]) == 0))
			deviceType = type_router;

		// Router
		else if (strcmp("--switch", argv[argLoop]) == 0)
			deviceType = type_switch;

		// Firewall
		else if ((strcmp("--firewall", argv[argLoop]) == 0) || (strcmp("--pix", argv[argLoop]) == 0) || (strcmp("--asa", argv[argLoop]) == 0) || (strcmp("--fwsm", argv[argLoop]) == 0))
			deviceType = type_firewall;

		// NMP
		else if ((strcmp("--catalyst", argv[argLoop]) == 0) || (strcmp("--nmp", argv[argLoop]) == 0))
			deviceType = type_catalyst;

		// Juniper ScreenOS
		else if ((strcmp("--netscreen", argv[argLoop]) == 0) || (strcmp("--screenos", argv[argLoop]) == 0))
			deviceType = type_screenos;

		// Set Timeout
		else if (strncmp("--timeout=", argv[argLoop], 10) == 0)
		{
			if (atoi(argv[argLoop] + 10) != 0)
				connectionTimeout = atoi(argv[argLoop] + 10);
		}

		// Dictionary supplied
		else if (strncmp("--css=", argv[argLoop], 6) == 0)
			cssFile = fopen(argv[argLoop] + 6, "r");

		// Company name specified
		else if (strncmp("--company-name=", argv[argLoop], 15) == 0)
			strncpy(companyName, argv[argLoop] + 15, sizeof(companyName) - 1);

		// Password Length
		else if (strncmp("--pass-length=", argv[argLoop], 14) == 0)
			passwordLength = atoi(argv[argLoop] + 14);

		// Password Uppers
		else if (strcmp("--pass-uppers=yes", argv[argLoop]) == 0)
			passwordUppers = true;
		else if (strcmp("--pass-uppers=no", argv[argLoop]) == 0)
			passwordUppers = false;

		// Password Lowers
		else if (strcmp("--pass-lowers=yes", argv[argLoop]) == 0)
			passwordLowers = true;
		else if (strcmp("--pass-lowers=no", argv[argLoop]) == 0)
			passwordLowers = false;

		// Password Either
		else if (strcmp("--pass-either=yes", argv[argLoop]) == 0)
			passwordEither = true;
		else if (strcmp("--pass-either=no", argv[argLoop]) == 0)
			passwordEither = false;

		// Password Numbers
		else if (strcmp("--pass-numbers=yes", argv[argLoop]) == 0)
			passwordNumbers = true;
		else if (strcmp("--pass-numbers=no", argv[argLoop]) == 0)
			passwordNumbers = false;

		// Password Specials
		else if (strcmp("--pass-specials=yes", argv[argLoop]) == 0)
			passwordSpecials = true;
		else if (strcmp("--pass-specials=no", argv[argLoop]) == 0)
			passwordSpecials = false;

		// Latex paper size
		else if (strncmp("--paper=", argv[argLoop], 8) == 0)
			strncpy(latex.paperSize, argv[argLoop]+8, sizeof(latex.paperSize) - 1);

		// Latex document class
		else if (strncmp("--documentclass=", argv[argLoop], 16) == 0)
			strncpy(latex.documentClass, argv[argLoop]+16, sizeof(latex.documentClass) - 1);

		// Appendix
		else if (strcmp("--no-appendix", argv[argLoop]) == 0)
			appendix = false;

		// Glossary
		else if (strcmp("--no-glossary", argv[argLoop]) == 0)
			glossary = false;

		// Common Ports
		else if (strcmp("--no-ports", argv[argLoop]) == 0)
			commonPorts = false;

		// Abbreviations
		else if (strcmp("--no-abbreviations", argv[argLoop]) == 0)
			abs = false;

		// Show host / network names instead of IP addresses (where possible)
		else if (strcmp("--no-names", argv[argLoop]) == 0)
			names = false;

		// Show host / network names instead of IP addresses (where possible)
		else if (strcmp("--debug-abbreviations", argv[argLoop]) == 0)
			debugAbbrev = true;

		// John the ripper output
		else if ((strncmp("--john=", argv[argLoop], 7) == 0) && (strlen(argv[argLoop]) > 7))
			strncpy(john, argv[argLoop]+7, sizeof(john) - 1);

		// Dictionary supplied
		else if (strncmp("--dictionary=", argv[argLoop], 13) == 0)
			dictionaryFile = fopen(argv[argLoop] + 13, "r");

		// Remote Cisco Device IP
		else if (strncmp("--cisco-ip=", argv[argLoop], 11) == 0)
			strncpy(remoteIP, argv[argLoop] + 11, sizeof(remoteIP) - 1);

		// Remote Cisco Device IP (Local)
		else if (strncmp("--local-ip=", argv[argLoop], 11) == 0)
			strncpy(localIP, argv[argLoop] + 11, sizeof(localIP) - 1);

		// Remote Cisco Device SNMP read/write community
		else if (strncmp("--snmp=", argv[argLoop], 7) == 0)
			strncpy(remoteSnmp, argv[argLoop] + 7, sizeof(remoteSnmp) - 1);

		// Local TFTP server to use
		else if (strncmp("--tftproot=", argv[argLoop], 11) == 0)
			strncpy(remoteTftp, argv[argLoop] + 11, sizeof(remoteTftp) - 1);

		// Local Cisco Config file to save to
		else if (strncmp("--cisco-file=", argv[argLoop], 13) == 0)
			localSave = argLoop;

		// Cisco retrival method (old or new)
		else if (strcmp("--cisco=old", argv[argLoop]) == 0)
			oldCisco = true;
		else if (strcmp("--cisco=new", argv[argLoop]) == 0)
			oldCisco = false;

		// If all else fails, display the help
		else
			showUsage = true;
	}

	// Show help?
	if (showUsage == true)
	{
		printf("%s%s\n\n%s", COL_BLUE, program_banner, RESET);
		printf("nipper is a Network Infrastructure Configuration Parser. nipper takes\n");
		printf("a network infrastructure device configuration, processes the file and\n");
		printf("details security-related issues with the configuration together with\n");
		printf("detailed recommendations. nipper was previous known as CiscoParse.\n\n");

		printf("nipper currently supports Cisco switches (IOS), Cisco Routers (IOS),\n");
        printf("Cisco Catalyst devices (IOS/CatOS/NMP), Cisco Firewalls (PIX/ASA/FWSM)\n");
		printf("and Juniper NetScreen (ScreenOS). Output from nipper is in HTML, Latex\n");
		printf("XML and Text. nipper will reverse any type-7 passwords, other encrypted\n");
		printf("passwords can be output to a John-the-Ripper file for further testing.\n");

		printf("By default, input is retrieved from stdin and is output (in HTML format)\n");
		printf("to stdout.\n\n");

		printf("%sCommand:%s\n\n", COL_BLUE, RESET);
		printf("    %s%s [Options]%s\n\n", COL_GREEN, argv[0], RESET);

		printf("%sGeneral Options:%s\n\n", COL_BLUE, RESET);
		printf("    %s--help%s\n", COL_GREEN, RESET);
		printf("    Displays this command line help. Overrides all other options.\n\n");
		printf("    %s--input=<file>%s\n", COL_GREEN, RESET);
		printf("    Specifies an input file for the Cisco configuration. If not specified\n");
		printf("    Nipper will get its configuration from stdin.\n\n");

		printf("%sNetwork Infrastructure Device Type Options:%s\n\n", COL_BLUE, RESET);
		printf("    Different types of network infrastructure devices have different\n");
		printf("    configuration file structures and default and options.\n\n");
		printf("    %s--firewall\n", COL_GREEN);
		printf("    --pix\n");
		printf("    --asa\n");
		printf("    --fwsm%s\n", RESET);
		printf("    Treat input configuration as a PIX/ASA/FWSM-based Cisco firewall.\n\n");
		printf("    %s--router\n", COL_GREEN);
		printf("    --ios%s\n", RESET);
		printf("    Treat input configuration as a IOS-based Cisco router. This is the\n");
		printf("    default device type.\n\n");
		printf("    %s--switch%s\n", COL_GREEN, RESET);
		printf("    Treat input configuration as a IOS-based Cisco switch or Catalyst.\n\n");
		printf("    %s--catalyst\n", COL_GREEN);
		printf("    --nmp%s\n", RESET);
		printf("    Treat input configuration as a NMP- or CatOS-based Cisco Catalyst.\n\n");
		printf("    %s--netscreen\n", COL_GREEN);
		printf("    --screenos%s\n", RESET);
		printf("    Treat input configuration as a ScreenOS-based Juniper Netscreen\n");
		printf("    device.\n\n");

		printf("%sGeneral Report Options:%s\n\n", COL_BLUE, RESET);
		printf("    %s--output=<file>\n", COL_GREEN);
		printf("    --report=<file>%s\n", RESET);
		printf("    Specified an output file for the parsed Cisco configuration. If not\n");
		printf("    specified Nipper will output to stdout.\n\n");
		printf("    %s--output-format=<html | latex | text | xml>\n", COL_GREEN);
		printf("    --report-format=<html | latex | text | xml>\n");
		printf("    --html | --latex | --text | --xml%s\n", RESET);
		printf("    Defines the output format produced by Nipper. If not specified\n");
		printf("    then output will be in HTML report format. Supported formats are \n");
		printf("    HTML, Latex and text.\n\n");
		printf("    %s--no-audit%s\n", COL_GREEN, RESET);
		printf("    This option disables the security audit of the device configuration.\n");
		printf("    By default a security audit is performed and the results, together\n");
		printf("    with recommendations, are included in the report output.\n\n");
		printf("    %s--css=<file>%s\n", COL_GREEN, RESET);
		printf("    Supply an external Cascading Style Sheet (CSS) for use with the HTML\n");
		printf("    -based reports. If this option is not supplied a default CSS is\n");
		printf("    embeded. The supplied CSS file will be embeded within the report\n");
		printf("    output.\n\n");
		printf("    %s--paper=<Latex paper size>%s\n", COL_GREEN, RESET);
		printf("    This option can be used to configure the Latex document paper size.\n");
		printf("    The default paper size is '%s'. Refer to the Latex documentation\n", latex.paperSize);
		printf("    for other settings.\n\n");
		printf("    %s--documentclass=<Latex document class>%s\n", COL_GREEN, RESET);
		printf("    This option can be used to configure the Latex document class. The\n");
		printf("    default document class is '%s'. Refer to the Latex documentation\n", latex.documentClass);
		printf("    for other settings.\n\n");
		printf("    %s--company-name=<name>%s\n", COL_GREEN, RESET);
		printf("    Outputs an alternative to 'Nipper recommends...' in the report\n");
		printf("    output. So for the Acme company, it would read, 'Acme recommends...'.\n\n");

		printf("%sAppendix Report Options:%s\n\n", COL_BLUE, RESET);
		printf("    %s--no-appendix%s\n", COL_GREEN, RESET);
		printf("    Disables the output of the report appendix section. By default an\n");
		printf("    appendix section is generated and added to the report.\n\n");
		printf("    %s--no-abbreviations%s\n", COL_GREEN, RESET);
		printf("    Disables the output of an abbreviations section in the appendix. By\n");
		printf("    default an abbreviations section is generated and added to the\n");
		printf("    appendix.\n\n");
		//printf("    %s--no-glossary%s\n", COL_GREEN, RESET);
		//printf("    Disables the output of a glossary section in the appendix. By default\n");
		//printf("    a glossary section is added to the appendix.\n\n");
		printf("    %s--no-ports%s\n", COL_GREEN, RESET);
		printf("    Disables the output of a common ports section in the appendix. By\n");
		printf("    default a common ports section is added to appendix.\n\n");

		printf("%sPassword Options:%s\n\n", COL_BLUE, RESET);
		printf("    Password strength and dictionary-based testing is performed on all\n");
		printf("    authentication systems. Therefore, in addition to user passwords,\n");
		printf("    SNMP community strings and keys are also checked.\n\n");
		printf("    %s--no-passwords%s\n", COL_GREEN, RESET);
		printf("    Removes passwords from the output. However, passwords are always\n");
		printf("    checked against defaults and for password strength.\n\n");
		printf("    %s--john=<file>%s\n", COL_GREEN, RESET);
		printf("    Output the Cisco type 5 passwords in a file that can then be run\n");
		printf("    through john-the-ripper.\n\n");
		printf("    %s--dictionary=<file>%s\n", COL_GREEN, RESET);
		printf("    Use a dictionary file to check the passwords and keys are not\n");
		printf("    dictionary-based words. If this option is not specified a small\n");
		printf("    number of built-in passwords are used instead. This option is\n");
		printf("    not used to break MD5 passwords, only compare passwords already\n");
		printf("    known.\n\n");
		printf("    %s--pass-length=<length>%s\n", COL_GREEN, RESET);
		printf("    Password strength checks will check that retrieved passwords have\n");
		printf("    a minimum length as specified. The default is %d characters.\n\n", passwordLength);
		printf("    %s--pass-uppers=<yes|no>%s\n", COL_GREEN, RESET);
		printf("    Password strength checks will check that retrieved passwords must\n");
		printf("    include uppercase characters. The default is ");
		if (passwordUppers == false)
			printf("no.\n\n");
		else
			printf("yes.\n\n");
		printf("    %s--pass-lowers=<yes|no>%s\n", COL_GREEN, RESET);
		printf("    Password strength checks will check that retrieved passwords must\n");
		printf("    include lowercase characters. The default is ");
		if (passwordLowers == false)
			printf("no.\n\n");
		else
			printf("yes.\n\n");
		printf("    %s--pass-either=<yes|no>%s\n", COL_GREEN, RESET);
		printf("    Password strength checks will check that retrieved passwords must\n");
		printf("    include upper or lowercase characters. The default is ");
		if (passwordEither == false)
			printf("no.\n\n");
		else
			printf("yes.\n\n");
		printf("    %s--pass-numbers=<yes|no>%s\n", COL_GREEN, RESET);
		printf("    Password strength checks will check that retrieved passwords must\n");
		printf("    include numbers. The default is ");
		if (passwordNumbers == false)
			printf("no.\n\n");
		else
			printf("yes.\n\n");
		printf("    %s--pass-specials=<yes|no>%s\n", COL_GREEN, RESET);
		printf("    Password strength checks will check that retrieved passwords must\n");
		printf("    include special characters. The default is ");
		if (passwordSpecials == false)
			printf("no.\n\n");
		else
			printf("yes.\n\n");

		printf("%sMisc Options:%s\n\n", COL_BLUE, RESET);
		printf("    %s--timeout=<seconds>%s\n", COL_GREEN, RESET);
		printf("    Specify the number of seconds that connection timeouts should be\n");
		printf("    less than or equal to. If timeouts are less than this, a security\n");
		printf("    will be generated in the report. The default is %d minutes.\n\n", (connectionTimeout / 60));
		printf("    %s--no-names%s\n", COL_GREEN, RESET);
		printf("    Convert host / network names to IP addresses where possible.\n\n");
		printf("    %s--expand-acl%s\n", COL_GREEN, RESET);
		printf("    Expand an Access Control Lists object definitions. This option will\n");
		printf("    show the full details of an ACL.\n\n");

		printf("%sRemote SNMP Config Retrival Options:%s\n\n", COL_BLUE, RESET);
		printf("    These options allow the retrieval of a Cisco configuration file from\n");
		printf("    the device itself. There are two methods of remote retrival; the first\n");
		printf("    uses the old Cisco system MIB which works with devices upto IOS 12.x,\n");
		printf("    the second method uses the new Cisco config copy MIB and works with\n");
		printf("    devices running IOS 12.0 (some with 11.2P). Cisco Parse defaults to\n");
		printf("    using the new Cisco system MIB.\n");
		printf("    This feature requires a running TFTP server and the net-snmp package.\n");
		printf("    Additionally, the IP address of the Cisco device and the local address\n");
		printf("    need to be specified. You will need to be the root user.\n\n");
		printf("    %s--cisco-ip=<ip address>%s\n", COL_GREEN, RESET);
		printf("    Specify the IP address of the Cisco Device to retrieve the config from.\n\n");
		printf("    %s--local-ip=<ip address>%s\n", COL_GREEN, RESET);
		printf("    The IP address of the host making the request. This is essential for\n");
		printf("    remote config retrival, else it will fail.\n\n");
		printf("    %s--cisco=<old | new>%s\n", COL_GREEN, RESET);
		printf("    There are two methods of retrieving the config remotely using SNMP. The\n");
		printf("    old method is supported on Cisco devices upto IOS 12.x, the\n");
		printf("    new method (default) is supported from IOS 12.0 (11.2P on some devices).\n\n");
		printf("    %s--snmp=<community>%s\n", COL_GREEN, RESET);
		printf("    The snmp community to be used with the above. A default of 'private'\n");
		printf("    will be used if not specified.\n\n");
		printf("    %s--tftproot=<directory>%s\n", COL_GREEN, RESET);
		printf("    The tftp root directory to be used for the Cisco config transfer. If\n");
		printf("    not specified a default of /tftproot/ will be used.\n\n");
		printf("    %s--cisco-file=<filename>%s\n", COL_GREEN, RESET);
		printf("    Specifies a filename for the Cisco config file to be saved to. If not\n");
		printf("    specified, then the file will be left in the tftproot as a random file\n");
		printf("    name.\n\n");

		printf("%sDebug Options:%s\n\n", COL_BLUE, RESET);
		printf("    %s--debug%s\n", COL_GREEN, RESET);
		printf("    Outputs additional information that most users would not find very\n");
		printf("    helpful, but is useful in understanding how the configuration is\n");
		printf("    interpreted. Using the debug option overrides all other output.\n\n");
		printf("    %s--debug-abbreviations%s\n", COL_GREEN, RESET);
		printf("    Shows a list of all the abbreviations that were not processed. The\n");
		printf("    list is output to stdout and lists the number of hits each entry\n");
		printf("    has had. Not all of the items listed will be valid abbreviations.\n\n");
		printf("    %s--debug-lnp%s\n", COL_GREEN, RESET);
		printf("    Output the configuration lines not processed (to screen only).\n\n");

		printf("%sExamples:%s\n\n", COL_BLUE, RESET);
		printf("    %s%s --pix --input=configIOS.txt --output=report.html%s\n", COL_GREEN, argv[0], RESET);
		printf("    Parses the PIX configuration file configIOS.txt and outputs the\n");
		printf("    the HTML report to the report.html file.\n\n");
		printf("    %scat ios.config | %s --company-name=Acme >report.html%s\n", COL_GREEN, argv[0], RESET);
		printf("    Parses the Cisco router configuration stored in ios.config and\n");
		printf("    redirects the HTML report output to report.html. Nipper will\n");
		printf("    also use the company name \"Acme\" in the report output.\n\n");
		printf("    %s%s --pass-length=12 --input=cisco.txt --output=report.htm%s\n", COL_GREEN, argv[0], RESET);
		printf("    Parses the configuration file stored in cisco.txt and outputs the\n");
		printf("    report to report.htm. It will override the default password length\n");
		printf("    check of %d and checks for a miniumum password length of 12 chars.\n\n", passwordLength);
	}


// ----------------------------------------------------
// PART 2: Process Configuration File

	else
	{
		// Debug Output?
		if (debugMode == true)
		{
			printf("%s\n", program_banner);

			printf("\n%sCommand Line Options\n====================%s\nCommand:", COL_BLUE, RESET);
			for (argLoop = 0; argLoop < argc; argLoop++)
			{
				printf(" %s", argv[argLoop]);
			}
			printf("\n\n");
		}

		// Get remote config?
		if ((remoteIP[0] != 0) && (localIP[0] != 0) && (getuid() == 0))
		{

			// Create tftp file
			tempInt = rand();
			if (remoteTftp[strlen(remoteTftp) -1] != '/')
				remoteTftp[strlen(remoteTftp)] = '/';
			sprintf(tempCommand, "touch %s%d.config", remoteTftp, tempInt);
			system(tempCommand);

			// Set tftp file writable
			sprintf(tempCommand, "chmod 666 %s%d.config", remoteTftp, tempInt);
			system(tempCommand);

			// Old method?
			if (oldCisco == true)
			{
				// http://www.cisco.com/en/US/tech/tk648/tk362/technologies_tech_note09186a008009463e.shtml
				// Write to SNMP server
				sprintf(tempCommand, "snmpset -c %s -v 1 %s 1.3.6.1.4.1.9.2.1.55.%s s \"%d.config\"", remoteSnmp, remoteIP, localIP, tempInt);
				if (debugMode == true)
					printf("SNMP Command: %s\n", tempCommand);
				system(tempCommand);
			}

			// new method
			else
			{
				// http://www.cisco.com/en/US/tech/tk648/tk362/technologies_tech_note09186a0080094aa6.shtml
				// Write to SNMP server
				tempInt2 = rand();
				sprintf(tempCommand, "snmpset -c %s -v 1 %s .1.3.6.1.4.1.9.9.96.1.1.1.1.2.%d i 1 .1.3.6.1.4.1.9.9.96.1.1.1.1.3.%d i 4 .1.3.6.1.4.1.9.9.96.1.1.1.1.4.%d i 1 .1.3.6.1.4.1.9.9.96.1.1.1.1.5.%d a \"%s\" .1.3.6.1.4.1.9.9.96.1.1.1.1.6.%d s \"%d.config\" .1.3.6.1.4.1.9.9.96.1.1.1.1.14.%d i 4", remoteSnmp, remoteIP, tempInt2, tempInt2, tempInt2, tempInt2, localIP, tempInt2, tempInt, tempInt2);
				if (debugMode == true)
					printf("SNMP Command: %s\n", tempCommand);
				system(tempCommand);
			}

			// Wait for a couple of secs (for file to arrive)
			system("sleep 3");

			// Check the file has content
			fileStats = malloc(sizeof(struct stat));
			memset(tempCommand, 0, sizeof(tempCommand));
			sprintf(tempCommand, "%s%d.config", remoteTftp, tempInt);
			stat(tempCommand, fileStats);

			// Cleanup SNMP if new...
			if (oldCisco == false)
			{
				memset(tempCommand, 0, sizeof(tempCommand));
				sprintf(tempCommand, "snmpset -c %s -v 1 %s .1.3.6.1.4.1.9.9.96.1.1.1.1.14.%d i 6", remoteSnmp, remoteIP, tempInt2);
				system(tempCommand);
			}

			// Exit if file has nothing in it...
			if (fileStats->st_size == 0)
			{
				free(fileStats);
				unlink(tempCommand);
				printf("\n%sERROR: Could not retrieve the config remotely.%s\n", COL_RED, RESET);
				return 1;
			}
			free(fileStats);

			// If file is to called something...
			if (localSave != 0)
			{
				memset(tempCommand, 0, sizeof(tempCommand));
				sprintf(tempCommand, "mv %s%d.config %s", remoteTftp, tempInt, argv[localSave]+13);
				system(tempCommand);
				memset(tempCommand, 0, sizeof(tempCommand));
				sprintf(tempCommand, "%s", argv[localSave]+13);
			}

			// Open file as input
			input = fopen(tempCommand, "r");
			if (input == NULL)
			{
				printf("\n%sERROR: Could not open the remote config file.%s\n", COL_RED, RESET);
				return 1;
			}
		}
		else if ((remoteIP[0] != 0) && (localIP[0] != 0) && (getuid() != 0))
		{
			printf("\n%sERROR: You must be the root user to get the Cisco config remotely.%s\n", COL_RED, RESET);
			return 1;
		}
		else if ((remoteIP[0] != 0) || (localIP[0] != 0))
		{
			printf("\n%sERROR: You must be the specify Cisco device and local IP.%s\n", COL_RED, RESET);
			return 1;
		}

		// From Stdin...
		else if (inputFile == 0)
			input = stdin;

		// From file...
		else
		{
			// Check the file has some contents...
			fileStats = malloc(sizeof(struct stat));
			stat(argv[inputFile]+8, fileStats);
			if (fileStats->st_size == 0)
			{
				free(fileStats);
				printf("\n%sERROR: The configuration file is empty!.%s\n", COL_RED, RESET);
				return 1;
			}
			free(fileStats);

			input = fopen(argv[inputFile]+8, "r");
			if (input == NULL)
			{
				printf("\n%sERROR: Could not open input file '%s'.%s\n", COL_RED, argv[inputFile]+8, RESET);
				return 1;
			}
		}

		if (debugMode == true)
		{
			printf("\n%sConfig Lines Examined\n=====================%s\n", COL_BLUE, RESET);
		}

		// Read in input and process...
		while (feof(input) == 0)
		{
			// Read line
			readLine(input, line, sizeof(line));

			switch (deviceType)
			{
				// IOS devices
				case type_router:
				case type_switch:
					if (strncmp("!", line, 1) == 0)
					{ }

					// Hostname
					else if (strncmp("hostname ", line, 9) == 0)
					{
						if (debugMode == true)
							printf("Hostname Line: %s\n", line);
						strncpy(configIOS.hostname, line+9, sizeof(configIOS.hostname) - 1);
					}

					// Version
					else if (strncmp("version ", line, 8) == 0)
					{
						if (debugMode == true)
							printf("Version Line: %s\n", line);
						strncpy(configIOS.version, line + 8, sizeof(configIOS.version) - 1);
						configIOS.versionMajor = atoi(configIOS.version);
						configIOS.versionMinor = atoi(strchr(configIOS.version, '.') + 1);
					}

					// Enable Passwords
					else if (strncmp("enable ", line, 7) == 0)
						processEnable(line, &configIOS, debugMode, dictionaryFile);

					// Usernames...
					else if (strncmp("username ", line, 9) == 0)
						processUsername(line, &configIOS, debugMode, dictionaryFile);

					// Service Password Encryption...
					else if (strncmp("service password-encryption", line, 27) == 0)
					{
						if (debugMode == true)
							printf("Service Password-Encryption Line: %s\n", line);
						configIOS.servicePasswordEncryption = true;
					}

					// Interfaces
					else if (strncmp("interface ", line, 10) == 0)
						processInterface(line, &configIOS, input, sizeof(line), debugMode, dictionaryFile);

					// Banner...
					else if (strncmp("banner ", line, 7) == 0)
						processBanner(line, &configIOS, input, sizeof(line), debugMode);

					// BootNetwork
					else if ((strncmp("no boot network", line, 15) == 0) || (strncmp("boot network", line, 12) == 0))
						processBootNetwork(line, &configIOS, debugMode);

					// SSH
					else if ((strncmp("no ip ssh", line, 9) == 0) || (strncmp("ip ssh", line, 6) == 0))
						processSsh(line, &configIOS, debugMode);

					// Service Config
					else if ((strncmp("no service config", line, 17) == 0) || (strncmp("service config", line, 14) == 0))
						processServiceConfig(line, &configIOS, debugMode);

					// Domain Lookup
					else if ((strncmp("no ip domain-lookup", line, 19) == 0) || (strncmp("ip domain-lookup", line, 16) == 0))
						processDomainLookup(line, &configIOS, debugMode);

					// Source routing...
					else if (strncmp("no ip source-route", line, 18) == 0)
					{
						if (debugMode == true)
							printf("IP Source Route Line: %s\n", line);
						configIOS.sourceRouting = false;
					}

					// PAD Service...
					else if (strncmp("no service pad", line, 14) == 0)
					{
						if (debugMode == true)
							printf("PAD Line: %s\n", line);
						configIOS.pad = false;
					}

					// Bootp...
					else if ((strncmp("no ip bootp server", line, 18) == 0) || (strncmp("ip bootp server", line, 15) == 0))
						processBootp(line, &configIOS, debugMode);

					// HTTP...
					else if ((strncmp("no ip http", line, 10) == 0) || (strncmp("ip http", line, 7) == 0))
						processHttp(line, &configIOS, debugMode);

					// TCP Small Servers...
					else if ((strncmp("no service tcp-small-servers", line, 28) == 0) || (strncmp("service tcp-small-servers", line, 25) == 0))
						processSmallServers(line, &configIOS, debugMode);

					// UDP Small Servers...
					else if ((strncmp("no service udp-small-servers", line, 28) == 0) || (strncmp("service udp-small-servers", line, 25) == 0))
						processSmallServers(line, &configIOS, debugMode);

					// Finger...
					else if ((strncmp("no service finger", line, 17) == 0) || (strncmp("no ip finger", line, 12) == 0))
					{
						if (debugMode == true)
							printf("Finger Line: %s\n", line);
						configIOS.finger = false;
					}

					// CDP...
					else if (strncmp("no cdp run", line, 10) == 0)
					{
						if (debugMode == true)
							printf("CDP Line: %s\n", line);
						configIOS.cdp = false;
					}

					// Logging...
					else if ((strncmp("logging ", line, 8) == 0) || (strncmp("no logging ", line, 11) == 0))
						processLogging(line, &configIOS, debugMode);

					// NTP...
					else if (strncmp("ntp ", line, 4) == 0)
						processNtp(line, &configIOS, debugMode, dictionaryFile);

					// Privilage Levels...
					else if (strncmp("privilage exec level ", line, 21) == 0)
						processPrivilage(line, &configIOS, debugMode);

					// SNMP...
					else if ((strncmp("snmp-server ", line, 12) == 0) || (strncmp("no snmp-server", line, 14) == 0))
						processSnmp(line, &configIOS, debugMode, dictionaryFile);

					// Access Lists...
					else if (strncmp("access-list ", line, 12) == 0)
						processAccessList(line, &configIOS, sizeof(line), debugMode);

					// FTP username / password...
					else if (strncmp("ip ftp ", line, 7) == 0)
						processFtp(line, &configIOS, debugMode, dictionaryFile);

					// Line...
					else if (strncmp("line ", line, 5) == 0)
						processLine(line, &configIOS, input, sizeof(line), debugMode, dictionaryFile);

					// Key chain...
					else if (strncmp("key chain ", line, 10) == 0)
						processKeyChain(line, &configIOS, input, sizeof(line), debugMode, dictionaryFile);

					// Routing...
					else if (strncmp("router ", line, 7) == 0)
						processRouter(line, &configIOS, input, sizeof(line), debugMode, dictionaryFile);

					// IP Route...
					else if (strncmp("ip route ", line, 9) == 0)
						processRoute(line, &configIOS, sizeof(line), debugMode);

					// TACACS...
					else if (strncmp("tacacs-server ", line, 14) == 0)
						processTacacs(line, &configIOS, debugMode, dictionaryFile);

					// AAA...
					else if (strncmp("aaa ", line, 4) == 0)
						processAaa(line, &configIOS, input, sizeof(line), debugMode, dictionaryFile);

					// Debug (lines not processed)
					else if (linesnotprocessed == true)
						printf("%s\n", line);
					break;


				// NMP devices
				case type_catalyst:

					// Hostname
					if (strncmp("set system name ", line, 16) == 0)
					{
						if (debugMode == true)
							printf("Hostname Line: %s\n", line);
						command = splitLine(line);
						strncpy(configNMP.hostname, command.part[3], sizeof(configNMP.hostname) - 1);
					}

					// Version
					else if (strncmp("#version ", line, 9) == 0)
					{
						if (debugMode == true)
							printf("Version Line: %s\n", line);
						strncpy(configNMP.version, line + 9, sizeof(configNMP.version) - 1);
						configNMP.versionMajor = atoi(configNMP.version);
						configNMP.versionMinor = atoi(strchr(configNMP.version, '.') + 1);
						configNMP.versionRevision = atoi(strchr(configNMP.version, '(') + 1);
					}

					// Logout Timeout...
					else if (strncmp("set logout ", line, 11) == 0)
					{
						if (debugMode == true)
							printf("Logout Line: %s\n", line);
						tempInt = 11;
						while (line[tempInt] == ' ')
							tempInt++;
						configNMP.timeout = atoi(line + tempInt);
					}

					// Location...
					else if (strncmp("set system location ", line, 20) == 0)
					{
						if (debugMode == true)
							printf("Location Line: %s\n", line);
						tempInt = 20;
						while (line[tempInt] == ' ')
							tempInt++;
						strncpy(configNMP.location, line + tempInt, sizeof(configNMP.location) - 1);
					}

					// Contact...
					else if (strncmp("set system contact ", line, 19) == 0)
					{
						if (debugMode == true)
							printf("Contact Line: %s\n", line);
						tempInt = 19;
						while (line[tempInt] == ' ')
							tempInt++;
						strncpy(configNMP.contact, line + tempInt, sizeof(configNMP.contact) - 1);
					}

					// Corefile...
					else if (strncmp("set system core-file ", line, 21) == 0)
					{
						if (debugMode == true)
							printf("Core-file Line: %s\n", line);
						tempInt = 21;
						while (line[tempInt] == ' ')
							tempInt++;
						strncpy(configNMP.coreFile, line + tempInt, sizeof(configNMP.coreFile) - 1);
					}

					// Syslog file...
					else if (strncmp("set system syslog-file ", line, 23) == 0)
					{
						if (debugMode == true)
							printf("Syslog-file Line: %s\n", line);
						tempInt = 23;
						while (line[tempInt] == ' ')
							tempInt++;
						strncpy(configNMP.syslogFile, line + tempInt, sizeof(configNMP.syslogFile) - 1);
					}

					// Enable Passwords
					else if ((strncmp("set password ", line, 13) == 0) || (strncmp("set enablepass ", line, 15) == 0))
						processEnablePass(line, &configNMP, debugMode);

					// CDP...
					else if (strncmp("set cdp ", line, 8) == 0)
						processNMPCDP(line, &configNMP, debugMode);

					// Permit List
					else if (strncmp("set ip permit ", line, 14) == 0)
						processNMPPermit(line, &configNMP, debugMode);

					// SNMP...
					else if (strncmp("set snmp ", line, 9) == 0)
						processNMPSnmp(line, &configNMP, debugMode, dictionaryFile);

					// Ports
					else if (strncmp("set port ", line, 9) == 0)
						processNMPPort(line, &configNMP, debugMode);

					// Trunk
					else if (strncmp("set trunk ", line, 10) == 0)
						processNMPTrunk(line, &configNMP, debugMode);

					// Spanning Tree
					else if (strncmp("set spantree ", line, 13) == 0)
						processNMPSpantree(line, &configNMP, debugMode);

					// VLAN
					else if (strncmp("set vlan ", line, 9) == 0)
						processNMPVlan(line, &configNMP, debugMode);

					// ICMP IP Unreachables
					else if (strncmp("set ip unreachable ", line, 19) == 0)
						processNMPUnreachable(line, &configNMP, debugMode);

					// IP Redirects
					else if (strncmp("set ip redirect ", line, 16) == 0)
						processNMPRedirect(line, &configNMP, debugMode);

					// IP Fragmentation
					else if (strncmp("set ip fragmentation ", line, 21) == 0)
						processNMPFragmentation(line, &configNMP, debugMode);

					// Debug (lines not processed)
					else if (linesnotprocessed == true)
						printf("%s\n", line);
					break;


				// PIX devices
				case type_firewall:

					// Hostname
					if (strncmp("hostname ", line, 9) == 0)
					{
						if (debugMode == true)
							printf("Hostname Line: %s\n", line);
						strncpy(configPIX.hostname, line+9, sizeof(configPIX.hostname) - 1);
					}

					// Version
					else if ((strncmp("PIX Version ", line, 12) == 0) || (strncmp("ASA Version ", line, 12) == 0) || (strncmp("FWSM Version ", line, 13) == 0))
					{
						if (debugMode == true)
							printf("Version Line: %s\n", line);
						strncpy(configPIX.version, line + 12, sizeof(configPIX.version) - 1);
						configPIX.versionMajor = atoi(configPIX.version);
						configPIX.versionMinor = atoi(strchr(configPIX.version, '.') + 1);
						configPIX.versionRevision = atoi(strchr(configPIX.version, '(') + 1);
						if (strncmp("PIX ", line, 4) == 0)
							configPIX.type = firewall_type_pix;
						else if (strncmp("ASA ", line, 4) == 0)
							configPIX.type = firewall_type_asa;
						else if (strncmp("FWSM ", line, 5) == 0)
							configPIX.type = firewall_type_fwsm;
					}

					// Transparent Firewall?
					else if (strncmp("firewall transparent ", line, 21) == 0)
					{
						if (debugMode == true)
							printf("Transparent Line: %s\n", line);
						configPIX.transparent = true;
					}

					// ICMP Access...
					else if (strncmp("icmp ", line, 5) == 0)
						processICMPPIX(line, &configPIX, sizeof(line), debugMode);

					// Enable Password
					else if ((strncmp("enable password ", line, 16) == 0) || (strncmp("passwd ", line, 7) == 0))
						processEnablePIX(line, &configPIX, debugMode);

					// Usernames...
					else if (strncmp("username ", line, 9) == 0)
						processPIXUsername(line, &configPIX, debugMode);

					// Route...
					else if (strncmp("route ", line, 6) == 0)
						processRoutePIX(line, &configPIX, debugMode);

					// Interfaces...
					else if ((strncmp("interface ", line, 10) == 0) || (strncmp("nameif ", line, 7) == 0) || (strncmp("mtu ", line, 4) == 0) || (strncmp("ip address ", line, 11) == 0) || (strncmp("access-group ", line, 13) == 0))
						processPIXInterface(line, &configPIX, input, sizeof(line), debugMode);

					// Object Group...
					else if (strncmp("object-group ", line, 13) == 0)
						processObjectGroup(line, &configPIX, input, sizeof(line), debugMode);

					// Protocol Inspection...
					else if ((strncmp("fixup protocol ", line, 15) == 0) || (strncmp("no fixup protocol ", line, 18) == 0))
						processFixupPIX(line, &configPIX, sizeof(line), debugMode);

					// Name...
					else if (strncmp("name ", line, 5) == 0)
						processPIXName(line, &configPIX, debugMode);

					// Domain Name
					else if (strncmp("domain-name ", line, 12) == 0)
						strncpy(configPIX.domainName, line + 12, sizeof(configPIX.domainName) - 1);

					// Access Lists...
					else if (strncmp("access-list ", line, 12) == 0)
						processPIXAccessList(line, &configPIX, sizeof(line), debugMode);

					// Static NAT / PAT
					else if (strncmp("static ", line, 7) == 0)
						processStaticPIX(line, &configPIX, sizeof(line), debugMode);

					// SSH
					else if (strncmp("ssh ", line, 4) == 0)
						processPIXSSH(line, &configPIX, sizeof(line), debugMode);

					// Debug (lines not processed)
					else if (linesnotprocessed == true)
						printf("%s\n", line);
					break;


				// ScreenOS devices
				case type_screenos:
					command = splitLine(line);
					if ((strcmp(command.part[0], "set") == 0) || (strcmp(command.part[0], "unset") == 0))
					{
						// Hostname
						if (strcmp(command.part[1], "hostname") == 0)
						{
							if (debugMode == true)
								printf("Hostname Line: %s\n", line);
							stripQuotes(command.part[2], line, sizeof(line));
							strncpy(configSOS.hostname, line, sizeof(configSOS.hostname) - 1);
						}

						// Admin
						else if (strcmp(command.part[1], "admin") == 0)
							processSOSAdmin(line, &configSOS, sizeof(line), debugMode);

						// Auth Server
						else if (strcmp(command.part[1], "auth-server") == 0)
							processSOSAuthServer(line, &configSOS, sizeof(line), debugMode);

						// Addresses
						else if (strcmp(command.part[1], "address") == 0)
							processSOSAddress(line, &configSOS, sizeof(line), debugMode);

						// Interface
						else if (strcmp(command.part[1], "interface") == 0)
							processSOSInterface(line, &configSOS, sizeof(line), debugMode);

						// Policy
						else if (strcmp(command.part[1], "policy") == 0)
							processSOSPolicy(line, &configSOS, input, sizeof(line), debugMode);

						// SSH
						else if (strcmp(command.part[1], "ssh") == 0)
							processSOSSSH(line, &configSOS, sizeof(line), debugMode);

						// Debug (lines not processed)
						else if (linesnotprocessed == true)
							printf("%s\n", line);
					}
					break;
			}
		}

		// Close input file...
		fclose(input);

		// Check that version has been identified for relevent configs...
		showUsage = false;
		switch (deviceType)
		{
			case type_switch:
			case type_router:
				if (configIOS.versionMajor == 0)
					showUsage = true;
				break;

			case type_catalyst:
				if (configNMP.versionMajor == 0)
					showUsage = true;
				break;

			case type_firewall:
				if (configPIX.versionMajor == 0)
					showUsage = true;
				break;

			case type_screenos:
				break;
		}
		if (showUsage == true)
		{
			printf("\n%sERROR: You are parsing the wrong type of configuration.\n       Specify the configuration type.%s\n", COL_RED, RESET);
			goto cleanup;
		}


// ----------------------------------------------------
// PART 3: Report Pre-parsing

		// Default Settings...
		switch(deviceType)
		{
			// Switch / Router / IOS
			case type_switch:
			case type_router:

				// Finger
				if (configIOS.finger == unconfigured)
				{
					if (((configIOS.versionMajor > 11) || ((configIOS.versionMajor == 11) && (configIOS.versionMinor > 2))) && ((configIOS.versionMajor < 12) || ((configIOS.versionMajor == 12) && (configIOS.versionMinor < 1))))
						configIOS.finger = true;
					else
						configIOS.finger = false;
				}

				// TCP Small Servers
				if (configIOS.tcpSmallServers == unconfigured)
				{
					if ((configIOS.versionMajor < 11) || ((configIOS.versionMajor == 11) && (configIOS.versionMinor < 3)))
						configIOS.tcpSmallServers = true;
					else
						configIOS.tcpSmallServers = false;
				}

				// UDP Small Servers
				if (configIOS.udpSmallServers == unconfigured)
				{
					if ((configIOS.versionMajor < 11) || ((configIOS.versionMajor == 11) && (configIOS.versionMinor < 3)))
						configIOS.udpSmallServers = true;
					else
						configIOS.udpSmallServers = false;
				}
				break;

			// Firewall / PIX
			case type_firewall:
				break;

			// Catalyst / NMP
			case type_catalyst:
				break;

			// ScreenOS
			case type_screenos:
				break;
		}

		// Convert Names to IP Addresses...
		if (names == false)
		{
			switch (deviceType)
			{
				case type_firewall:
					if (configPIX.nameList != 0)
					{
						// Convert Interfaces
						interfacePIXPointer = configPIX.interface;
						while (interfacePIXPointer != 0)
						{
							// Search for a match
							namePIXPointer = configPIX.nameList;
							while ((namePIXPointer->next != 0) && (strcmp(namePIXPointer->name, interfacePIXPointer->ipAddress) != 0))
								namePIXPointer = namePIXPointer->next;

							// If match has been found
							if (strcmp(namePIXPointer->name, interfacePIXPointer->ipAddress) == 0)
							{
								strncpy(interfacePIXPointer->ipAddress, namePIXPointer->address, sizeof(interfacePIXPointer->ipAddress) - 1);
								interfacePIXPointer->ipAddress[strlen(namePIXPointer->address)] = 0;
							}

							// Next Interface
							interfacePIXPointer = interfacePIXPointer->next;
						}

						// Static NAT (Mapped IP)
						staticPIXPointer = configPIX.staticNAT;
						while (staticPIXPointer != 0)
						{
							// Search for a match
							namePIXPointer = configPIX.nameList;
							while ((namePIXPointer->next != 0) && (strcmp(namePIXPointer->name, staticPIXPointer->ipMapped) != 0))
								namePIXPointer = namePIXPointer->next;

							// If match has been found
							if (strcmp(namePIXPointer->name, staticPIXPointer->ipMapped) == 0)
							{
								strncpy(staticPIXPointer->ipMapped, namePIXPointer->address, sizeof(staticPIXPointer->ipMapped) - 1);
								staticPIXPointer->ipMapped[strlen(namePIXPointer->address)] = 0;
							}

							// Next
							staticPIXPointer = staticPIXPointer->next;
						}

						// Static NAT (Real IP)
						staticPIXPointer = configPIX.staticNAT;
						while (staticPIXPointer != 0)
						{
							// Search for a match
							namePIXPointer = configPIX.nameList;
							while ((namePIXPointer->next != 0) && (strcmp(namePIXPointer->name, staticPIXPointer->ipReal) != 0))
								namePIXPointer = namePIXPointer->next;

							// If match has been found
							if (strcmp(namePIXPointer->name, staticPIXPointer->ipReal) == 0)
							{
								strncpy(staticPIXPointer->ipReal, namePIXPointer->address, sizeof(staticPIXPointer->ipReal) - 1);
								staticPIXPointer->ipReal[strlen(namePIXPointer->address)] = 0;
							}

							// Next
							staticPIXPointer = staticPIXPointer->next;
						}

						// Static Routing (IP Address)
						routePIXPointer = configPIX.route;
						while (routePIXPointer != 0)
						{
							// Search for a match
							namePIXPointer = configPIX.nameList;
							while ((namePIXPointer->next != 0) && (strcmp(namePIXPointer->name, routePIXPointer->ipAddress) != 0))
								namePIXPointer = namePIXPointer->next;

							// If match has been found
							if (strcmp(namePIXPointer->name, routePIXPointer->ipAddress) == 0)
							{
								strncpy(routePIXPointer->ipAddress, namePIXPointer->address, sizeof(routePIXPointer->ipAddress) - 1);
								routePIXPointer->ipAddress[strlen(namePIXPointer->address)] = 0;
							}

							// Next
							routePIXPointer = routePIXPointer->next;
						}

						// Static Routing (Gateway)
						routePIXPointer = configPIX.route;
						while (routePIXPointer != 0)
						{
							// Search for a match
							namePIXPointer = configPIX.nameList;
							while ((namePIXPointer->next != 0) && (strcmp(namePIXPointer->name, routePIXPointer->gateway) != 0))
								namePIXPointer = namePIXPointer->next;

							// If match has been found
							if (strcmp(namePIXPointer->name, routePIXPointer->gateway) == 0)
							{
								strncpy(routePIXPointer->gateway, namePIXPointer->address, sizeof(routePIXPointer->gateway) - 1);
								routePIXPointer->gateway[strlen(namePIXPointer->address)] = 0;
							}

							// Next
							routePIXPointer = routePIXPointer->next;
						}

						// ACL (Source)
						accessPIXPointer = configPIX.access;
						while (accessPIXPointer != 0)
						{
							permissionPIXPointer = accessPIXPointer->permission;
							while (permissionPIXPointer != 0)
							{
								// Search for a match
								namePIXPointer = configPIX.nameList;
								while ((namePIXPointer->next != 0) && (strcmp(namePIXPointer->name, permissionPIXPointer->source) != 0))
									namePIXPointer = namePIXPointer->next;

								// If match has been found
								if (strcmp(namePIXPointer->name, permissionPIXPointer->source) == 0)
								{
									strncpy(permissionPIXPointer->source, namePIXPointer->address, sizeof(permissionPIXPointer->source) - 1);
									permissionPIXPointer->source[strlen(namePIXPointer->address)] = 0;
								}

								// Next
								permissionPIXPointer = permissionPIXPointer->next;
							}

							// Next
							accessPIXPointer = accessPIXPointer->next;
						}

						// ACL (Destination)
						accessPIXPointer = configPIX.access;
						while (accessPIXPointer != 0)
						{
							permissionPIXPointer = accessPIXPointer->permission;
							while (permissionPIXPointer != 0)
							{
								// Search for a match
								namePIXPointer = configPIX.nameList;
								while ((namePIXPointer->next != 0) && (strcmp(namePIXPointer->name, permissionPIXPointer->dest) != 0))
									namePIXPointer = namePIXPointer->next;

								// If match has been found
								if (strcmp(namePIXPointer->name, permissionPIXPointer->dest) == 0)
								{
									strncpy(permissionPIXPointer->dest, namePIXPointer->address, sizeof(permissionPIXPointer->dest) - 1);
									permissionPIXPointer->dest[strlen(namePIXPointer->address)] = 0;
								}

								// Next
								permissionPIXPointer = permissionPIXPointer->next;
							}

							// Next
							accessPIXPointer = accessPIXPointer->next;
						}

						// Object Groups
						objectGroupPointer = configPIX.objectGroup;
						while (objectGroupPointer != 0)
						{
							if (objectGroupPointer->type == object_type_network)
							{
								objectPointer = objectGroupPointer->object;
								while (objectPointer != 0)
								{
									if ((objectPointer->type == object_type_host) || (objectPointer->type == object_type_network))
									{
										// Search for a match
										namePIXPointer = configPIX.nameList;
										while ((namePIXPointer->next != 0) && (strcmp(namePIXPointer->name, objectPointer->object) != 0))
											namePIXPointer = namePIXPointer->next;

										// If match has been found
										if (strcmp(namePIXPointer->name, objectPointer->object) == 0)
										{
											strncpy(objectPointer->object, namePIXPointer->address, sizeof(objectPointer->object) - 1);
											objectPointer->object[strlen(namePIXPointer->address)] = 0;
										}
									}

									// Next
									objectPointer = objectPointer->next;
								}
							}

							// Next
							objectGroupPointer = objectGroupPointer->next;
						}
					}
					break;

				case type_screenos:
					if (configSOS.nameList != 0)
					{
						// Convert Policy Rules
						policySOSPointer = configSOS.policies;
						while (policySOSPointer != 0)
						{
							ruleSOSPointer = policySOSPointer->policyRules;
							while (ruleSOSPointer != 0)
							{

								// Source Address
								// Find relevent zone name list
								addressSOSPointer = configSOS.nameList;
								while ((strcmp(policySOSPointer->fromZone, addressSOSPointer->zone) != 0) && (addressSOSPointer->next != 0))
									addressSOSPointer = addressSOSPointer->next;
								// If zone found
								if (strcmp(policySOSPointer->fromZone, addressSOSPointer->zone) == 0)
								{
									listSOSPointer = ruleSOSPointer->source;
									while (listSOSPointer != 0)
									{
										mappingSOSPointer = addressSOSPointer->mapping;
										while ((mappingSOSPointer->next != 0) && (strcmp(listSOSPointer->name, mappingSOSPointer->name) != 0))
											mappingSOSPointer = mappingSOSPointer->next;

										// If found
										if (strcmp(listSOSPointer->name, mappingSOSPointer->name) == 0)
										{
											if (mappingSOSPointer->netMask[0] == 0)
												sprintf(line, "%s", mappingSOSPointer->address);
											else
												sprintf(line, "%s %s", mappingSOSPointer->address, mappingSOSPointer->netMask);
											strncpy(listSOSPointer->name, line, sizeof(listSOSPointer->name) - 1);
										}
										listSOSPointer = listSOSPointer->next;
									}
								}

								// Destination Address
								// Find relevent zone name list
								addressSOSPointer = configSOS.nameList;
								while ((strcmp(policySOSPointer->toZone, addressSOSPointer->zone) != 0) && (addressSOSPointer->next != 0))
									addressSOSPointer = addressSOSPointer->next;
								// If zone found
								if (strcmp(policySOSPointer->toZone, addressSOSPointer->zone) == 0)
								{
									listSOSPointer = ruleSOSPointer->dest;
									while (listSOSPointer != 0)
									{
										mappingSOSPointer = addressSOSPointer->mapping;
										while ((mappingSOSPointer->next != 0) && (strcmp(listSOSPointer->name, mappingSOSPointer->name) != 0))
											mappingSOSPointer = mappingSOSPointer->next;

										// If found
										if (strcmp(listSOSPointer->name, mappingSOSPointer->name) == 0)
										{
											if (mappingSOSPointer->netMask[0] == 0)
												sprintf(line, "%s", mappingSOSPointer->address);
											else
												sprintf(line, "%s %s", mappingSOSPointer->address, mappingSOSPointer->netMask);
											strncpy(listSOSPointer->name, line, sizeof(listSOSPointer->name) - 1);
										}
										listSOSPointer = listSOSPointer->next;
									}
								}
								ruleSOSPointer = ruleSOSPointer->next;
							}
							policySOSPointer = policySOSPointer->next;
						}
					}
					break;
			}
		}


// ----------------------------------------------------
// PART 4: Reporting

		// Output debug information?
		if (debugMode == true)
		{
			switch (deviceType)
			{
				case type_router:
				case type_switch:
					reportIOSDebug(&configIOS);
					break;

				case type_firewall:
					reportPIXDebug(&configPIX);
					break;

				case type_catalyst:
					reportNMPDebug(&configNMP);
					break;

				case type_screenos:
					reportSOSDebug(&configSOS);
					break;
			}
		}

		// Output report & john (if specified)
		else
		{
			// Output john?
			int writeJohn = false;
			if ((deviceType == type_router) || (deviceType == type_switch))
			{
				if ((john[0] != 0) && (configIOS.john != 0))
					writeJohn = true;
			}
			else if (deviceType == type_firewall)
			{
				if ((john[0] != 0) && (configPIX.john != 0))
					writeJohn = true;
			}
			else if (deviceType == type_catalyst)
			{
				if ((john[0] != 0) && (configNMP.john != 0))
					writeJohn = true;
			}
			else if (deviceType == type_screenos)
			{
				if ((john[0] != 0) && (configSOS.john != 0))
					writeJohn = true;
			}
			if (writeJohn == true)
			{
				// Variables
				FILE *johnFile;

				// Create file
				johnFile = fopen(john, "w");
				if (johnFile == NULL)
				{
					printf("\n%sERROR: Could not create john-the-ripper file '%s'.%s\n", COL_RED, john, RESET);
					return 1;
				}

				// Write file contents
				if ((deviceType == type_router) || (deviceType == type_switch))
					johnPointer = configIOS.john;
				else if (deviceType == type_firewall)
					johnPointer = configPIX.john;
				else if (deviceType == type_catalyst)
					johnPointer = configNMP.john;
				else if (deviceType == type_screenos)
					johnPointer = configSOS.john;
				while (johnPointer != 0)
				{
					fprintf(johnFile, "%s:%s\n", johnPointer->username, johnPointer->password);
					johnPointer = johnPointer->next;
				}

				// Close file
				fclose(johnFile);
			}

			// Generate the report
			switch (deviceType)
			{
				case type_router:
				case type_switch:
					generateIOSReport(outputFormat, &configIOS, outputFile, sizeof(line), passwords, fullReports, cssFile, &latex, companyName, abs, appendix, glossary, commonPorts);
					break;

				case type_firewall:
					generatePIXReport(outputFormat, &configPIX, outputFile, sizeof(line), passwords, fullReports, cssFile, &latex, companyName, abs, appendix, glossary, commonPorts, expandACL);
					break;

				case type_catalyst:
					generateNMPReport(outputFormat, &configNMP, outputFile, sizeof(line), passwords, fullReports, cssFile, &latex, companyName, abs, appendix, glossary, commonPorts);
					break;

				case type_screenos:
					generateSOSReport(outputFormat, &configSOS, outputFile, sizeof(line), passwords, fullReports, cssFile, &latex, companyName, abs, appendix, glossary, commonPorts);
					break;
			}

			// Output Debug Abbreviations
			if (debugAbbrev == true)
			{
				printf("\n%sUnknown Abbreviations\n=====================%s\n", COL_BLUE, RESET);
				debugAbPointer = debugAb;
				while (debugAbPointer != 0)
				{
					printf("  %s (%d)\n", debugAbPointer->ab, debugAbPointer->count);
					debugAbPointer = debugAbPointer->next;
				}
			}
		}


// ----------------------------------------------------
// PART 5: Cleanup

cleanup:

		// Debug Abbreviations
		while (debugAb != 0)
		{
			debugAbPointer = debugAb->next;
			free (debugAb);
			debugAb = debugAbPointer;
		}

		if ((deviceType == type_router) || (deviceType == type_switch))
		{
			// Cleanup enable passwords
			while (configIOS.enable != 0)
			{
				enablePointer = configIOS.enable->next;
				free(configIOS.enable);
				configIOS.enable = enablePointer;
			}

			// Cleanup usernames
			while (configIOS.username != 0)
			{
				usernamePointer = configIOS.username->next;
				free(configIOS.username);
				configIOS.username = usernamePointer;
			}

			// Cleanup Privilages
			while (configIOS.privilage != 0)
			{
				privilagePointer = configIOS.privilage->next;
				free(configIOS.privilage);
				configIOS.privilage = privilagePointer;
			}

			// Cleanup Route
			while (configIOS.route != 0)
			{
				routePointer = configIOS.route->next;
				free(configIOS.route);
				configIOS.route = routePointer;
			}

			// Cleanup HTTP
			if (configIOS.http != 0)
				free(configIOS.http);

			// Cleanup Kay Chain
			while (configIOS.chain != 0)
			{
				while (configIOS.chain->key != 0)
				{
					keyPointer = configIOS.chain->key->next;
					free (configIOS.chain->key);
					configIOS.chain->key = keyPointer;
				}
				keyChainPointer = configIOS.chain->next;
				free (configIOS.chain);
				configIOS.chain = keyChainPointer;
			}

			// Cleanup SSH
			if (configIOS.ssh != 0)
			{
				while (configIOS.ssh->interface != 0)
				{
					interfacePointer = configIOS.ssh->interface->next;
					free(configIOS.ssh->interface);
					configIOS.ssh->interface = interfacePointer;
				}
				free(configIOS.ssh);
			}

			// Cleanup Line
			while (configIOS.line != 0)
			{
				linePointer = configIOS.line->next;
				free(configIOS.line);
				configIOS.line = linePointer;
			}

			// Cleanup BGP
			if (configIOS.bgp != 0)
			{
				while (configIOS.bgp->neighbor != 0)
				{
					neighborPointer = configIOS.bgp->neighbor->next;
					free(configIOS.bgp->neighbor);
					configIOS.bgp->neighbor = neighborPointer;
				}
				free(configIOS.bgp);
			}

			// Interfaces
			while (configIOS.interface != 0)
			{
				interfaceConfigPointer = configIOS.interface->next;
				free(configIOS.interface);
				configIOS.interface = interfaceConfigPointer;
			}

			// Banner
			while (configIOS.banner != 0)
			{
				bannerPointer = configIOS.banner->next;
				free(configIOS.banner->line);
				free(configIOS.banner);
				configIOS.banner = bannerPointer;
			}

			// Cleanup AAA
			if (configIOS.aaa != 0)
			{
				while (configIOS.aaa->groupServer != 0)
				{
					while (configIOS.aaa->groupServer->server != 0)
					{
						aaaServerPointer = configIOS.aaa->groupServer->server->next;
						free (configIOS.aaa->groupServer->server);
						configIOS.aaa->groupServer->server = aaaServerPointer;
					}
					aaaGroupServerPointer = configIOS.aaa->groupServer->next;
					free (configIOS.aaa->groupServer);
					configIOS.aaa->groupServer = aaaGroupServerPointer;
				}
				free (configIOS.aaa);
			}

			// Cleanup Logging
			if (configIOS.logging != 0)
			{
				while (configIOS.logging->logServer != 0)
				{
					logServerPointer = configIOS.logging->logServer->next;
					free(configIOS.logging->logServer);
					configIOS.logging->logServer = logServerPointer;
				}
				free(configIOS.logging);
			}

			// Cleanup TACACS
			if (configIOS.tacacs != 0)
			{
				while (configIOS.tacacs->host != 0)
				{
					tacacsHostPointer = configIOS.tacacs->host->next;
					free(configIOS.tacacs->host);
					configIOS.tacacs->host = tacacsHostPointer;
				}
				free(configIOS.tacacs);
			}

			// OSPF
			if (configIOS.ospf != 0)
			{
				// Cleanup OSPF Area
				while (configIOS.ospf->ospfArea != 0)
				{
					ospfAreaPointer = configIOS.ospf->ospfArea->next;
					free(configIOS.ospf->ospfArea);
					configIOS.ospf->ospfArea = ospfAreaPointer;
				}
				// Cleanup OSPF Network
				while (configIOS.ospf->ospfNetwork != 0)
				{
					ospfNetworkPointer = configIOS.ospf->ospfNetwork->next;
					free(configIOS.ospf->ospfNetwork);
					configIOS.ospf->ospfNetwork = ospfNetworkPointer;
				}
				// Cleanup OSPF Neighbor
				while (configIOS.ospf->neighbor != 0)
				{
					ospfNeighborPointer = configIOS.ospf->neighbor->next;
					free(configIOS.ospf->neighbor);
					configIOS.ospf->neighbor = ospfNeighborPointer;
				}
				free(configIOS.ospf);
			}

			// Cleanup ftp
			if (configIOS.ftp != 0)
			{
				while (configIOS.ftp->interface != 0)
				{
					interfacePointer = configIOS.ftp->interface->next;
					free(configIOS.ftp->interface);
					configIOS.ftp->interface = interfacePointer;
				}
				free(configIOS.ftp);
			}

			// Cleanup ntp
			if (configIOS.ntp != 0)
			{
				while (configIOS.ntp->key != 0)
				{
					ntpKeyPointer = configIOS.ntp->key->next;
					free(configIOS.ntp->key);
					configIOS.ntp->key = ntpKeyPointer;
				}
				while (configIOS.ntp->server != 0)
				{
					ntpHostPointer = configIOS.ntp->server->next;
					free(configIOS.ntp->server);
					configIOS.ntp->server = ntpHostPointer;
				}
				while (configIOS.ntp->peer != 0)
				{
					ntpHostPointer = configIOS.ntp->peer->next;
					free(configIOS.ntp->peer);
					configIOS.ntp->peer = ntpHostPointer;
				}
				while (configIOS.ntp->access != 0)
				{
					ntpAccessPointer = configIOS.ntp->access->next;
					free(configIOS.ntp->access);
					configIOS.ntp->access = ntpAccessPointer;
				}
				free(configIOS.ntp);
			}

			// Cleanup snmp
			if (configIOS.snmp != 0)
			{
				while (configIOS.snmp->community != 0)
				{
					communityPointer = configIOS.snmp->community->next;
					free(configIOS.snmp->community);
					configIOS.snmp->community = communityPointer;
				}
				while (configIOS.snmp->trap != 0)
				{
					snmpTrapPointer = configIOS.snmp->trap->next;
					free(configIOS.snmp->trap);
					configIOS.snmp->trap = snmpTrapPointer;
				}
				while (configIOS.snmp->host != 0)
				{
					snmpHostPointer = configIOS.snmp->host->next;
					free(configIOS.snmp->host);
					configIOS.snmp->host = snmpHostPointer;
				}
				while (configIOS.snmp->view != 0)
				{
					while (configIOS.snmp->view->viewItem != 0)
					{
						snmpViewItemPointer = configIOS.snmp->view->viewItem->next;
						free(configIOS.snmp->view->viewItem);
						configIOS.snmp->view->viewItem = snmpViewItemPointer;
					}
					snmpViewPointer = configIOS.snmp->view->next;
					free(configIOS.snmp->view);
					configIOS.snmp->view = snmpViewPointer;
				}
				while (configIOS.snmp->group != 0)
				{
					while (configIOS.snmp->group->user != 0)
					{
						snmpUserPointer = configIOS.snmp->group->user->next;
						free(configIOS.snmp->group->user);
						configIOS.snmp->group->user = snmpUserPointer;
					}
					snmpGroupPointer = configIOS.snmp->group->next;
					free(configIOS.snmp->group);
					configIOS.snmp->group = snmpGroupPointer;
				}
				free(configIOS.snmp);
			}

			// Cleanup access-lists
			while (configIOS.access != 0)
			{
				while (configIOS.access->permission != 0)
				{
					permissionPointer = configIOS.access->permission->next;
					free(configIOS.access->permission);
					configIOS.access->permission = permissionPointer;
				}
				accessPointer = configIOS.access->next;
				free(configIOS.access);
				configIOS.access = accessPointer;
			}

			// Cleanup John-the-ripper
			while (configIOS.john != 0)
			{
				johnPointer = configIOS.john->next;
				free(configIOS.john);
				configIOS.john = johnPointer;
			}
		}

		else if (deviceType == type_firewall)
		{
			// Cleanup Interfaces
			while (configPIX.interface != 0)
			{
				interfacePIXPointer = configPIX.interface->next;
				free(configPIX.interface);
				configPIX.interface = interfacePIXPointer;
			}

			// Cleanup Static NAT / PAT
			while (configPIX.staticNAT != 0)
			{
				staticPIXPointer = configPIX.staticNAT->next;
				free(configPIX.staticNAT);
				configPIX.staticNAT = staticPIXPointer;
			}

			// Cleanup SSH
			if (configPIX.ssh != 0)
				free(configPIX.ssh);

			// Cleanup access-lists
			while (configPIX.access != 0)
			{
				while (configPIX.access->permission != 0)
				{
					permissionPIXPointer = configPIX.access->permission->next;
					free(configPIX.access->permission);
					configPIX.access->permission = permissionPIXPointer;
				}
				accessPIXPointer = configPIX.access->next;
				free(configPIX.access);
				configPIX.access = accessPIXPointer;
			}

			// Cleanup Onjects
			while (configPIX.objectGroup != 0)
			{
				while (configPIX.objectGroup->object != 0)
				{
					objectPointer = configPIX.objectGroup->object->next;
					free(configPIX.objectGroup->object);
					configPIX.objectGroup->object = objectPointer;
				}
				objectGroupPointer = configPIX.objectGroup->next;
				free(configPIX.objectGroup);
				configPIX.objectGroup = objectGroupPointer;
			}

			// Free route
			while (configPIX.route != 0)
			{
				routePIXPointer = configPIX.route->next;
				free (configPIX.route);
				configPIX.route = routePIXPointer;
			}

			// Free name list
			while (configPIX.nameList != 0)
			{
				namePIXPointer = configPIX.nameList->next;
				free (configPIX.nameList);
				configPIX.nameList = namePIXPointer;
			}

			// Free Enable Passwords
			while (configPIX.enable != 0)
			{
				enablePIXPointer = configPIX.enable->next;
				free (configPIX.enable);
				configPIX.enable = enablePIXPointer;
			}

			// Free Usernames
			while (configPIX.username != 0)
			{
				usernamePIXPointer = configPIX.username->next;
				free (configPIX.username);
				configPIX.username = usernamePIXPointer;
			}

			// Fixup
			while (configPIX.fixup != 0)
			{
				fixupPIXPointer = configPIX.fixup->next;
				free (configPIX.fixup);
				configPIX.fixup = fixupPIXPointer;
			}

			// ICMP Access
			while (configPIX.icmp != 0)
			{
				icmpPIXPointer = configPIX.icmp->next;
				free (configPIX.icmp);
				configPIX.icmp = icmpPIXPointer;
			}

			// Cleanup John-the-ripper
			while (configPIX.john != 0)
			{
				johnPointer = configPIX.john->next;
				free(configPIX.john);
				configPIX.john = johnPointer;
			}
		}
		else if (deviceType == type_catalyst)
		{
			// Cleanup SNMP
			if (configNMP.snmp != 0)
			{
				while (configNMP.snmp->community != 0)
				{
					communityNMPPointer = configNMP.snmp->community->next;
					free(configNMP.snmp->community);
					configNMP.snmp->community = communityNMPPointer;
				}
				while (configNMP.snmp->trap != 0)
				{
					snmpTrapPointer = configNMP.snmp->trap->next;
					free(configNMP.snmp->trap);
					configNMP.snmp->trap = snmpTrapPointer;
				}
				while (configNMP.snmp->host != 0)
				{
					snmpNMPHostPointer = configNMP.snmp->host->next;
					free(configNMP.snmp->host);
					configNMP.snmp->host = snmpNMPHostPointer;
				}
				free(configNMP.snmp);
			}

			// Cleanup Modules & Ports
			while (configNMP.module != 0)
			{
				while (configNMP.module->ports != 0)
				{
					portNMPPointer = configNMP.module->ports->next;
					free(configNMP.module->ports);
					configNMP.module->ports = portNMPPointer;
				}
				moduleNMPPointer = configNMP.module->next;
				free(configNMP.module);
				configNMP.module = moduleNMPPointer;
			}

			// Cleanup VLANS
			while (configNMP.vlan != 0)
			{
				vlanNMPPointer = configNMP.vlan->next;
				free(configNMP.vlan);
				configNMP.vlan = vlanNMPPointer;
			}

			// Cleanup Permit Lists
			if (configNMP.permit != 0)
			{
				while (configNMP.permit->telnetHost != 0)
				{
					hostNMPPointer = configNMP.permit->telnetHost->next;
					free(configNMP.permit->telnetHost);
					configNMP.permit->telnetHost = hostNMPPointer;
				}
				while (configNMP.permit->sshHost != 0)
				{
					hostNMPPointer = configNMP.permit->sshHost->next;
					free(configNMP.permit->sshHost);
					configNMP.permit->sshHost = hostNMPPointer;
				}
				while (configNMP.permit->snmpHost != 0)
				{
					hostNMPPointer = configNMP.permit->snmpHost->next;
					free(configNMP.permit->snmpHost);
					configNMP.permit->snmpHost = hostNMPPointer;
				}
				free(configNMP.permit);
			}

			// Cleanup switch config
			if (configNMP.switchConfig != 0)
				free (configNMP.switchConfig);

			// Cleanup John-the-ripper
			while (configNMP.john != 0)
			{
				johnPointer = configNMP.john->next;
				free(configNMP.john);
				configNMP.john = johnPointer;
			}
		}
		else if (deviceType == type_screenos)
		{
			// Cleanup Policy
			while (configSOS.policies != 0)
			{
				// Cleanup rules
				while (configSOS.policies->policyRules != 0)
				{
					// Cleanup Source
					while (configSOS.policies->policyRules->source != 0)
					{
						listSOSPointer = configSOS.policies->policyRules->source->next;
						free (configSOS.policies->policyRules->source);
						configSOS.policies->policyRules->source = listSOSPointer;
					}
					// Cleanup Dest
					while (configSOS.policies->policyRules->dest != 0)
					{
						listSOSPointer = configSOS.policies->policyRules->dest->next;
						free (configSOS.policies->policyRules->dest);
						configSOS.policies->policyRules->dest = listSOSPointer;
					}
					// Cleanup Service
					while (configSOS.policies->policyRules->service != 0)
					{
						listSOSPointer = configSOS.policies->policyRules->service->next;
						free (configSOS.policies->policyRules->service);
						configSOS.policies->policyRules->service = listSOSPointer;
					}
					ruleSOSPointer = configSOS.policies->policyRules->next;
					free (configSOS.policies->policyRules);
					configSOS.policies->policyRules = ruleSOSPointer;
				}
				policySOSPointer = configSOS.policies->next;
				free (configSOS.policies);
				configSOS.policies = policySOSPointer;
			}

			// Cleanup Name List Address Mappings
			if (configSOS.nameList != 0)
			{
				// Cleanup Admin Users
				while (configSOS.nameList->mapping != 0)
				{
					mappingSOSPointer = configSOS.nameList->mapping->next;
					free(configSOS.nameList->mapping);
					configSOS.nameList->mapping = mappingSOSPointer;
				}
				addressSOSPointer = configSOS.nameList->next;
				free (configSOS.nameList);
				configSOS.nameList = addressSOSPointer;
			}

			// Cleanup admin
			if (configSOS.admin != 0)
			{
				// Cleanup Admin Users
				while (configSOS.admin->users != 0)
				{
					userSOSPointer = configSOS.admin->users->next;
					free(configSOS.admin->users);
					configSOS.admin->users = userSOSPointer;
				}
				free (configSOS.admin);

				// Cleanup auth server
				while (configSOS.authServer != 0)
				{
					authServerPointer = configSOS.authServer->next;
					free (configSOS.authServer);
					configSOS.authServer = authServerPointer;
				}

				// Cleanup Interfaces
				while (configSOS.interface != 0)
				{
					interfaceSOSPointer = configSOS.interface->next;
					free (configSOS.interface);
					configSOS.interface = interfaceSOSPointer;
				}
			}

			// Cleanup Auth Server
			if (configSOS.authServer != 0)
				free (configSOS.authServer);

			// Cleanup John-the-ripper
			while (configSOS.john != 0)
			{
				johnPointer = configSOS.john->next;
				free(configSOS.john);
				configSOS.john = johnPointer;
			}
		}

		// close dict file
		if (dictionaryFile != 0)
		{
			fclose(dictionaryFile);
		}

		// close html css file
		if (cssFile != 0)
		{
			fclose(cssFile);
		}
	}

	return 0;
}
